import argparse
import os
import pickle
from typing import List, Tuple, Dict

import numpy as np
import torch
import torch.nn.functional as F
from torch.utils.data import dataloader
from torchvision import transforms

from lofarnn.models.base.cnn import f1_loss
from lofarnn.models.base.resnet import BinaryFocalLoss
from lofarnn.models.dataloaders.datasets import RadioSourceDataset


def default_argument_parser():
    """
    Create a parser with some common arguments.

    Returns:
        argparse.ArgumentParser:
    """
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--resume",
        action="store_true",
        help="whether to attempt to resume from the checkpoint directory",
    )
    parser.add_argument(
        "--augment",
        action="store_true",
        help="whether to augment input data, default False",
    )
    parser.add_argument(
        "--norm",
        action="store_true",
        help="whether to normalize magnitudes or not, default False",
    )
    parser.add_argument(
        "--single",
        action="store_true",
        help="whether to use single source or multiple, default False",
    )
    parser.add_argument(
        "--shuffle",
        action="store_true",
        help="whether to use shuffle multisource order, no effect for single source, default False",
    )
    parser.add_argument(
        "--num-sources", type=int, default=40, help="max number of sources to include",
    )
    parser.add_argument(
        "--nodes", type=int, default=1, help="max number of nodes to use",
    )
    parser.add_argument(
        "-g", "--gpus", default=1, type=int, help="number of gpus per node"
    )
    parser.add_argument(
        "-nr", "--nr", default=0, type=int, help="ranking within the nodes"
    )
    parser.add_argument(
        "--num-trials", type=int, default=100, help="max number of trials to perform",
    )
    parser.add_argument(
        "--classes", type=int, default=40, help="max number of sources to include",
    )
    parser.add_argument(
        "--dataset", type=str, default="", help="path to dataset annotations files"
    )
    parser.add_argument(
        "--loss",
        type=str,
        default="cross-entropy",
        help="loss to use, from 'cross-entropy' (default), 'focal', 'f1' ",
    )
    parser.add_argument("--experiment", type=str, default="", help="experiment name")
    parser.add_argument("--lr", type=float, default=0.001, help="learning rate")
    parser.add_argument(
        "--lr-type",
        type=str,
        default="",
        help="learning rate type: None (default), 'plataeu', or 'cyclical' ",
    )
    parser.add_argument("--batch", type=int, default=32, help="batch size")
    parser.add_argument("--fraction", type=float, default=1.0, help="batch size")
    parser.add_argument("--epochs", type=int, default=200, help="number of epochs")
    parser.add_argument(
        "--log-interval",
        type=int,
        default=20,
        help="number of minibatches between logging",
    )

    return parser


def only_image_transforms(
    image: np.ndarray, sources: List[str]
) -> Tuple[np.ndarray, List[str]]:
    """
    Only applies transforms to the image, and leaves the sources as they are
    """
    sequence = transforms.Compose(
        [
            transforms.ToPILImage(),
            transforms.RandomVerticalFlip(),
            transforms.RandomHorizontalFlip(),
            transforms.RandomRotation(180),
            transforms.ToTensor(),
            transforms.RandomErasing(),
        ]
    )
    return sequence(image), sources


def setup(args) -> Tuple[RadioSourceDataset, RadioSourceDataset, RadioSourceDataset]:
    """
    Setup dataset and dataloaders for these new datasets
    """
    train_dataset = RadioSourceDataset(
        [
            os.path.join(args.dataset, f"cnn_train_test_norm{args.norm}_extra.pkl"),
            os.path.join(args.dataset, f"cnn_val_norm{args.norm}_extra.pkl"),
        ],
        single_source_per_img=args.single,
        shuffle=args.shuffle,
        norm=not args.norm,
        num_sources=args.num_sources,
        fraction=args.fraction,
        transform=only_image_transforms if args.augment else None,
    )
    train_test_dataset = RadioSourceDataset(
        [
            os.path.join(args.dataset, f"cnn_train_test_norm{args.norm}_extra.pkl"),
            os.path.join(args.dataset, f"cnn_val_norm{args.norm}_extra.pkl"),
        ],
        single_source_per_img=args.single,
        shuffle=args.shuffle,
        norm=not args.norm,
        num_sources=args.num_sources,
    )
    val_dataset = RadioSourceDataset(
        os.path.join(args.dataset, f"cnn_test_norm{args.norm}_extra.pkl"),
        single_source_per_img=False,
        shuffle=args.shuffle,
        norm=not args.norm,
        num_sources=args.num_sources,
    )
    args.world_size = args.gpus * args.nodes
    return train_dataset, train_test_dataset, val_dataset


def test(
    args,
    model: torch.nn.Module,
    device: torch.device,
    test_loader: dataloader,
    epoch: int,
    name: str = "Test",
    output_dir: str = "./",
    config: Dict[str, str] = {"loss": "cross-entropy"},
) -> float:
    save_test_loss = []
    save_correct = []
    save_recalls = []
    recall = 0

    named_recalls = {}

    model.eval()
    test_loss = 0
    correct = 0
    loss_fn = BinaryFocalLoss(
        alpha=[config["alpha_1"], config["alpha_2"]],
        gamma=config["gamma"],
        reduction="mean",
    )
    with torch.no_grad():
        for data in test_loader:
            image, source, labels, names = (
                data["images"].to(device),
                data["sources"].to(device),
                data["labels"].to(device),
                data["names"],
            )
            output = model(image, source)
            # sum up batch loss
            if config["loss"] == "cross-entropy":
                try:
                    test_loss += F.binary_cross_entropy(
                        F.softmax(output, dim=-1), labels
                    ).item()
                except RuntimeError:
                    print(output)
            elif config["loss"] == "f1":
                test_loss += f1_loss(
                    output, labels.argmax(dim=1), is_training=False
                ).item()
            elif config["loss"] == "focal":
                test_loss += loss_fn(output, labels).item()
            # get the index of the max log-probability
            pred = output.argmax(dim=1, keepdim=True)
            label = labels.argmax(dim=1, keepdim=True)
            correct += pred.eq(label.view_as(pred)).sum().item()

            # Now get named recall ones
            if not args.single:
                for i in range(len(names)):
                    # Assumes testing is with batch size of 1
                    named_recalls[names[i]] = pred.eq(label.view_as(pred)).sum().item()
                    recall += pred.eq(label.view_as(pred)).sum().item()
            else:
                for i in range(len(names)):
                    if (
                        label.item() == 0
                    ):  # Label is source, don't care about the many negative examples
                        if pred.item() == 0:  # Prediction is source
                            named_recalls[names[i]] = 1  # Value is correct
                            recall += 1
                        else:  # Prediction is not correct
                            named_recalls[names[i]] = 0  # Value is incorrect

            save_test_loss.append(test_loss)
            save_correct.append(correct)
    recall /= len(test_loader.dataset.annotations)
    save_recalls.append(recall)
    test_loss /= len(test_loader)

    print(
        "\n{} set: Average loss: {:.4f}, Accuracy: {}/{} ({:.0f}%) Recall: {:.2f}%\n".format(
            name,
            test_loss,
            correct,
            len(test_loader),
            100.0 * correct / len(test_loader),
            100.0 * recall,
        )
    )
    pickle.dump(
        named_recalls,
        open(os.path.join(output_dir, f"{name}_source_recall_epoch{epoch}.pkl"), "wb"),
    )
    a = np.asarray(save_test_loss)
    with open(os.path.join(output_dir, f"{name}_loss.csv"), "ab") as f:
        np.savetxt(f, a, delimiter=",")
    a = np.asarray(save_recalls)
    with open(os.path.join(output_dir, f"{name}_recall.csv"), "ab") as f:
        np.savetxt(f, a, delimiter=",")
    if config["single"]:
        return correct
    else:
        return test_loss


def train(
    args,
    model: torch.nn.Module,
    device: torch.device,
    train_loader: dataloader,
    optimizer,
    scheduler,
    epoch: int,
    output_dir: str = "./",
    config: Dict[str, str] = {"loss": "cross-entropy"},
) -> None:
    save_loss = []
    total_loss = 0
    model.train()
    loss_fn = BinaryFocalLoss(
        alpha=[config["alpha_1"], config["alpha_2"]],
        gamma=config["gamma"],
        reduction="mean",
    )
    for batch_idx, data in enumerate(train_loader):
        image, source, labels, names = (
            data["images"].to(device),
            data["sources"].to(device),
            data["labels"].to(device),
            data["names"],
        )
        optimizer.zero_grad()
        output = model(image, source)
        if config["loss"] == "cross-entropy":
            loss = F.binary_cross_entropy(F.softmax(output, dim=-1), labels)
        elif config["loss"] == "f1":
            loss = f1_loss(output, labels.argmax(dim=1), is_training=True)
        elif config["loss"] == "focal":
            loss = loss_fn(output, labels)
        else:
            raise Exception("Loss not one of 'cross-entropy', 'focal', 'f1' ")
        loss.backward()

        save_loss.append(loss.item())

        optimizer.step()

        if args.lr_type == "cyclical":
            scheduler.step()
        total_loss += loss.item()
        if batch_idx % args.log_interval == 0:
            print(
                "Train Epoch: {}\tLoss: {:.6f} \t Average loss {:.6f}".format(
                    epoch, loss.item(), np.mean(save_loss[-args.log_interval :])
                )
            )
    if args.lr_type == "plateau":
        scheduler.step(total_loss)  # LROnPlateau step is after each epoch
    a = np.asarray(save_loss)
    with open(os.path.join(output_dir, "train_loss.csv"), "ab") as f:
        np.savetxt(f, a, delimiter=",")



import torch
import torch.nn as nn
import torch.nn.functional as F

from lofarnn.models.base.resnet import ResNet, Bottleneck

"""

Taken from https://discuss.pytorch.org/t/concatenating-observations-that-include-image-pose-and-sensor-readings/41084

This is for having a CNN look at the radio data and generate feature maps, and then in the last step of the classifier,
goes through and adds in the optical source features, like distance from radio center, multispectral magnitudes, etc.

"""


class RadioSingleSourceModel(nn.Module):
    """
    Model for classifying whether a single optical source is the radio source's source
    """

    def __init__(self, num_image_layers, len_aux_data, config):
        super(RadioSingleSourceModel, self).__init__()
        self.config = config

        # Image net
        self.cnn = ResNet(
            num_image_layers,
            Bottleneck,
            [3, 4, 23, 3],
            num_classes=2,
            zero_init_residual=False,
            groups=32,
            width_per_group=8,
            keep_fc=True,
        )
        self.cnn.fc = nn.Linear(self.cnn.fc.in_features, self.config["fc_out"])

        # Source net
        self.source1 = nn.Linear(len_aux_data, self.config["fc_final"])
        self.source2 = nn.Linear(self.config["fc_final"], self.config["fc_out"])

        # Combined
        self.fc1 = nn.Linear(
            self.config["fc_out"] + self.config["fc_out"], self.config["fc_final"]
        )
        self.fc2 = nn.Linear(self.config["fc_final"], 2)

    def forward(self, image, data):
        x1 = self.cnn(image)
        if self.config["act"] == "leaky":
            x2 = F.leaky_relu(self.source1(data))
        elif self.config["act"] == "elu":
            x2 = F.elu(self.source1(data))
        else:
            x2 = F.relu(self.source1(data))
        x2 = self.source2(x2)

        # Combine them
        x = torch.cat((x1, x2), dim=1)
        if self.config["act"] == "leaky":
            x = F.leaky_relu(self.fc1(x))
        elif self.config["act"] == "elu":
            x = F.elu(self.fc1(x))
        else:
            x = F.relu(self.fc1(x))
        x = self.fc2(x)
        return x


class RadioMultiSourceModel(nn.Module):
    """
    Model for classifying which of a set number of sources is the optical source
    """

    def __init__(self, num_image_layers, num_sources, config):
        super(RadioMultiSourceModel, self).__init__()
        self.config = config

        # Image net - ResNet 101 32x8d
        self.cnn = ResNet(
            num_image_layers,
            Bottleneck,
            [3, 4, 23, 3],
            num_classes=num_sources,
            zero_init_residual=False,
            groups=32,
            width_per_group=8,
            keep_fc=True,
        )
        self.cnn.fc = nn.Linear(self.cnn.fc.in_features, self.config["fc_out"])

        # Source net - smaller ResNet-50 to hold aux-data x num_sources 'image' to classify which of those souces is the source
        self.cnn2 = ResNet(
            num_image_layers,
            Bottleneck,
            [3, 4, 6, 3],
            num_classes=num_sources,
            keep_fc=True,
        )
        self.cnn2.fc = nn.Linear(self.cnn2.fc.in_features, self.config["fc_out"])

        # Combined
        self.fc1 = nn.Linear(
            self.config["fc_out"] + self.config["fc_out"], self.config["fc_final"]
        )
        self.fc2 = nn.Linear(self.config["fc_final"], num_sources)

    def forward(self, image, data):
        x1 = self.cnn(image)
        x2 = self.cnn2(data)

        # Combine them
        x = torch.cat((x1, x2), dim=1)
        if self.config["act"] == "leaky":
            x = F.leaky_relu(self.fc1(x))
        elif self.config["act"] == "elu":
            x = F.elu(self.fc1(x))
        else:
            x = F.relu(self.fc1(x))
        x = self.fc2(x)
        return x


class RadioMultiSourceEnsembleModel(nn.Module):
    """
    Model for classifying which of a set number of sources is the optical source
    Uses an Ensemble of multiple networks per input to give the output
    """

    def __init__(self, num_image_layers, num_sources, config):
        super(RadioMultiSourceEnsembleModel, self).__init__()
        self.config = config

        # Image net - ResNet 101 32x8d
        self.cnn = ResNet(
            num_image_layers,
            Bottleneck,
            [3, 4, 23, 3],
            num_classes=num_sources,
            zero_init_residual=False,
            groups=32,
            width_per_group=8,
            keep_fc=True,
        )
        self.cnn.fc = nn.Linear(self.cnn.fc.in_features, self.config["fc_out"])

        # Source net - ResNet 101 32x8d
        self.cnn2 = ResNet(
            num_image_layers,
            Bottleneck,
            [3, 4, 23, 3],
            num_classes=num_sources,
            zero_init_residual=False,
            groups=32,
            width_per_group=8,
            keep_fc=True,
        )
        self.cnn2.fc = nn.Linear(self.cnn2.fc.in_features, self.config["fc_out"])

        # Combined
        self.fc1 = nn.Linear(
            self.config["fc_out"] + self.config["fc_out"], self.config["fc_final"]
        )
        self.fc2 = nn.Linear(self.config["fc_final"], num_sources)

    def forward(self, image, data):
        x1 = self.cnn(image)
        x2 = self.cnn2(data)

        # Combine them
        x = torch.cat((x1, x2), dim=1)
        if self.config["act"] == "leaky":
            x = F.leaky_relu(self.fc1(x))
        elif self.config["act"] == "elu":
            x = F.elu(self.fc1(x))
        else:
            x = F.relu(self.fc1(x))
        x = self.fc2(x)
        return x


def f1_loss(
    y_pred: torch.Tensor, y_true: torch.Tensor, is_training=False
) -> torch.Tensor:
    """Calculate F1 score. Can work with gpu tensors

    The original implmentation is written by Michal Haltuf on Kaggle.

    Returns
    -------
    torch.Tensor
        `ndim` == 1. 0 <= val <= 1

    Reference
    ---------
    - https://www.kaggle.com/rejpalcz/best-loss-function-for-f1-score-metric
    - https://scikit-learn.org/stable/modules/generated/sklearn.metrics.f1_score.html#sklearn.metrics.f1_score
    - https://discuss.pytorch.org/t/calculating-precision-recall-and-f1-score-in-case-of-multi-label-classification/28265/6

    """
    assert y_true.ndim == 1
    assert y_pred.ndim == 1 or y_pred.ndim == 2

    if y_pred.ndim == 2:
        y_pred = y_pred.argmax(dim=1)

    tp = (y_true * y_pred).sum().to(torch.float32)
    fp = ((1 - y_true) * y_pred).sum().to(torch.float32)
    fn = (y_true * (1 - y_pred)).sum().to(torch.float32)

    epsilon = 1e-7

    precision = tp / (tp + fp + epsilon)
    recall = tp / (tp + fn + epsilon)

    f1 = 2 * (precision * recall) / (precision + recall + epsilon)
    f1.requires_grad = is_training
    return f1


class F1_Loss(nn.Module):
    """Calculate F1 score. Can work with gpu tensors

    The original implmentation is written by Michal Haltuf on Kaggle.

    Returns
    -------
    torch.Tensor
        `ndim` == 1. epsilon <= val <= 1

    Reference
    ---------
    - https://www.kaggle.com/rejpalcz/best-loss-function-for-f1-score-metric
    - https://scikit-learn.org/stable/modules/generated/sklearn.metrics.f1_score.html#sklearn.metrics.f1_score
    - https://discuss.pytorch.org/t/calculating-precision-recall-and-f1-score-in-case-of-multi-label-classification/28265/6
    - http://www.ryanzhang.info/python/writing-your-own-loss-function-module-for-pytorch/
    """

    def __init__(self, epsilon=1e-7):
        super().__init__()
        self.epsilon = epsilon

    def forward(self, y_pred, y_true):
        assert y_pred.ndim == 2
        assert y_true.ndim == 1
        y_true = F.one_hot(y_true, 2).to(torch.float32)
        y_pred = F.softmax(y_pred, dim=1)

        tp = (y_true * y_pred).sum(dim=0).to(torch.float32)
        fp = ((1 - y_true) * y_pred).sum(dim=0).to(torch.float32)
        fn = (y_true * (1 - y_pred.item())).sum(dim=0).to(torch.float32)

        precision = tp / (tp + fp + self.epsilon)
        recall = tp / (tp + fn + self.epsilon)

        f1 = 2 * (precision * recall) / (precision + recall + self.epsilon)
        f1 = f1.clamp(min=self.epsilon, max=1 - self.epsilon)
        return 1 - f1.mean().item()



from typing import Tuple, List

import matplotlib.patches as patches
import matplotlib.pyplot as plt
import numpy as np


def plot_cutout_and_bboxes(image_name: str, title: str) -> Tuple[plt.Figure, plt.Axes]:
    """

    :param image_name:
    :param title:
    :return:
    """
    print(image_name)
    cutout_and_bboxes = np.load(image_name, allow_pickle=True)
    cutout = cutout_and_bboxes[0]
    bboxes = cutout_and_bboxes[1]
    fig, ax = plt.subplots(1)
    if cutout.shape[2] == 3:
        ax.imshow(cutout)

    for bbox in bboxes:
        rect = patches.Rectangle(
            (int(bbox[0]), int(bbox[1])),
            1,
            1,
            linewidth=1,
            edgecolor="w",
            facecolor="none",
        )
        ax.add_patch(rect)
    plt.title(title)
    return fig, ax


def plot_three_channel_debug(
    image_stack: np.ndarray,
    cutouts: List[Tuple[float]],
    scale_size: float,
    source_loc: Tuple[float, float],
) -> None:
    """
    Plots the 3 channel image in debug
    """
    # Plot
    fig, axs = plt.subplots(2, 2)
    image_stack = np.asarray(image_stack)
    print(image_stack.shape)
    image_stack = image_stack[:, :, :3]  # Take first three

    cax_00 = axs[0, 0].imshow(image_stack)
    rect = patches.Rectangle(
        (float(cutouts[0][0]), float(cutouts[0][1])),
        float(cutouts[0][2]) - float(cutouts[0][0]),
        float(cutouts[0][3]) - float(cutouts[0][1]),
        linewidth=1,
        edgecolor="r",
        facecolor="none",
    )
    axs[0, 0].add_patch(rect)
    axs[0, 0].scatter(
        source_loc[1],
        source_loc[0],
        s=scale_size,
        edgecolor="black",
        facecolor=(1, 1, 1, 0.15),
    )
    axs[0, 0].xaxis.set_major_formatter(plt.NullFormatter())  # kill xlabels
    axs[0, 0].yaxis.set_major_formatter(plt.NullFormatter())  # kill ylabels
    axs[0, 0].set_title("Combined")

    cax_01 = axs[0, 1].imshow(image_stack[:, :, 0], cmap="Reds")
    fig.colorbar(cax_01, ax=axs[0, 1])
    rect = patches.Rectangle(
        (float(cutouts[0][0]), float(cutouts[0][1])),
        float(cutouts[0][2]) - float(cutouts[0][0]),
        float(cutouts[0][3]) - float(cutouts[0][1]),
        linewidth=1,
        edgecolor="r",
        facecolor="none",
    )
    axs[0, 1].add_patch(rect)
    axs[0, 1].scatter(
        source_loc[1],
        source_loc[0],
        s=scale_size,
        edgecolor="black",
        facecolor=(1, 1, 1, 0.15),
    )
    axs[0, 1].xaxis.set_major_formatter(plt.NullFormatter())
    axs[0, 1].yaxis.set_major_formatter(plt.NullFormatter())
    axs[0, 1].set_title("LOFAR")

    cax_10 = axs[1, 0].imshow(image_stack[:, :, 1], cmap="Greens")
    fig.colorbar(cax_10, ax=axs[1, 0])
    rect = patches.Rectangle(
        (float(cutouts[0][0]), float(cutouts[0][1])),
        float(cutouts[0][2]) - float(cutouts[0][0]),
        float(cutouts[0][3]) - float(cutouts[0][1]),
        linewidth=1,
        edgecolor="r",
        facecolor="none",
    )
    axs[1, 0].add_patch(rect)
    axs[1, 0].scatter(
        source_loc[1],
        source_loc[0],
        s=scale_size,
        edgecolor="black",
        facecolor=(1, 1, 1, 0.15),
    )
    axs[1, 0].xaxis.set_major_formatter(plt.NullFormatter())
    axs[1, 0].yaxis.set_major_formatter(plt.NullFormatter())
    axs[1, 0].set_title("i Band")

    cax_11 = axs[1, 1].imshow(image_stack[:, :, 2], cmap="Blues")
    fig.colorbar(cax_11, ax=axs[1, 1])
    rect = patches.Rectangle(
        (float(cutouts[0][0]), float(cutouts[0][1])),
        float(cutouts[0][2]) - float(cutouts[0][0]),
        float(cutouts[0][3]) - float(cutouts[0][1]),
        linewidth=1,
        edgecolor="r",
        facecolor="none",
    )
    axs[1, 1].add_patch(rect)
    axs[1, 1].scatter(
        source_loc[1],
        source_loc[0],
        s=scale_size,
        edgecolor="black",
        facecolor=(1, 1, 1, 0.15),
    )
    axs[1, 1].xaxis.set_major_formatter(plt.NullFormatter())
    axs[1, 1].yaxis.set_major_formatter(plt.NullFormatter())
    axs[1, 1].set_title("W1 Band")
    plt.show()



import torch.nn as nn
from antialiased_cnns import BlurPool

"""
Antialiased ResNet
"""


def conv3x3(in_planes, out_planes, stride=1, groups=1):
    """3x3 convolution with padding"""
    return nn.Conv2d(
        in_planes,
        out_planes,
        kernel_size=3,
        stride=stride,
        padding=1,
        groups=groups,
        bias=False,
    )


def conv1x1(in_planes, out_planes, stride=1):
    """1x1 convolution"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=1, stride=stride, bias=False)


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(
        self,
        inplanes,
        planes,
        stride=1,
        downsample=None,
        groups=1,
        norm_layer=None,
        filter_size=1,
    ):
        super(BasicBlock, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if groups != 1:
            raise ValueError("BasicBlock only supports groups=1")
        # Both self.conv1 and self.downsample layers downsample the input when stride != 1
        self.conv1 = conv3x3(inplanes, planes)
        self.bn1 = norm_layer(planes)
        self.relu = nn.ReLU(inplace=True)
        if stride == 1:
            self.conv2 = conv3x3(planes, planes)
        else:
            self.conv2 = nn.Sequential(
                BlurPool(planes, filt_size=filter_size, stride=stride),
                conv3x3(planes, planes),
            )
        self.bn2 = norm_layer(planes)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out


class Bottleneck(nn.Module):
    expansion = 4

    def __init__(
        self,
        inplanes,
        planes,
        stride=1,
        downsample=None,
        groups=1,
        norm_layer=None,
        filter_size=1,
    ):
        super(Bottleneck, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        # Both self.conv2 and self.downsample layers downsample the input when stride != 1
        self.conv1 = conv1x1(inplanes, planes)
        self.bn1 = norm_layer(planes)
        self.conv2 = conv3x3(planes, planes, groups)  # stride moved
        self.bn2 = norm_layer(planes)
        if stride == 1:
            self.conv3 = conv1x1(planes, planes * self.expansion)
        else:
            self.conv3 = nn.Sequential(
                BlurPool(planes, filt_size=filter_size, stride=stride),
                conv1x1(planes, planes * self.expansion),
            )
        self.bn3 = norm_layer(planes * self.expansion)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out


class AntiAliasedResNet(nn.Module):
    def __init__(
        self,
        num_channels,
        block,
        layers,
        num_classes=2,
        zero_init_residual=False,
        groups=1,
        width_per_group=64,
        norm_layer=None,
        filter_size=1,
        pool_only=True,
        keep_fc=True,
    ):
        super(AntiAliasedResNet, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        planes = [int(width_per_group * groups * 2 ** i) for i in range(4)]
        self.inplanes = 64
        self.keep_fc = keep_fc

        if pool_only:
            self.conv1 = nn.Conv2d(
                num_channels,
                self.inplanes,
                kernel_size=7,
                stride=2,
                padding=3,
                bias=False,
            )
        else:
            self.conv1 = nn.Conv2d(
                num_channels,
                self.inplanes,
                kernel_size=7,
                stride=1,
                padding=3,
                bias=False,
            )
        self.bn1 = norm_layer(self.inplanes)
        self.relu = nn.ReLU(inplace=True)

        if pool_only:
            self.maxpool = nn.Sequential(
                *[
                    nn.MaxPool2d(kernel_size=2, stride=1),
                    BlurPool(planes[0], filt_size=filter_size, stride=2,),
                ]
            )
        else:
            self.maxpool = nn.Sequential(
                *[
                    BlurPool(planes[0], filt_size=filter_size, stride=2,),
                    nn.MaxPool2d(kernel_size=2, stride=1),
                    BlurPool(planes[0], filt_size=filter_size, stride=2,),
                ]
            )

        self.layer1 = self._make_layer(
            block, 64, layers[0], groups=groups, norm_layer=norm_layer
        )
        self.layer2 = self._make_layer(
            block,
            128,
            layers[1],
            stride=2,
            groups=groups,
            norm_layer=norm_layer,
            filter_size=filter_size,
        )
        self.layer3 = self._make_layer(
            block,
            256,
            layers[2],
            stride=2,
            groups=groups,
            norm_layer=norm_layer,
            filter_size=filter_size,
        )
        self.layer4 = self._make_layer(
            block,
            512,
            layers[3],
            stride=2,
            groups=groups,
            norm_layer=norm_layer,
            filter_size=filter_size,
        )
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = nn.Linear(512 * block.expansion, num_classes)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                if (
                    m.in_channels != m.out_channels
                    or m.out_channels != m.groups
                    or m.bias is not None
                ):
                    # don't want to reinitialize downsample layers, code assuming normal conv layers will not have these characteristics
                    nn.init.kaiming_normal_(
                        m.weight, mode="fan_out", nonlinearity="relu"
                    )
                else:
                    print("Not initializing")
            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

        # Zero-initialize the last BN in each residual branch,
        # so that the residual branch starts with zeros, and each residual block behaves like an identity.
        # This improves the model by 0.2~0.3% according to https://arxiv.org/abs/1706.02677
        if zero_init_residual:
            for m in self.modules():
                if isinstance(m, Bottleneck):
                    nn.init.constant_(m.bn3.weight, 0)
                elif isinstance(m, BasicBlock):
                    nn.init.constant_(m.bn2.weight, 0)

    def _make_layer(
        self, block, planes, blocks, stride=1, groups=1, norm_layer=None, filter_size=1
    ):
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            # downsample = nn.Sequential(
            #     conv1x1(self.inplanes, planes * block.expansion, stride, filter_size=filter_size),
            #     norm_layer(planes * block.expansion),
            # )

            downsample = (
                [
                    BlurPool(
                        filt_size=filter_size, stride=stride, channels=self.inplanes
                    ),
                ]
                if (stride != 1)
                else []
            )
            downsample += [
                conv1x1(self.inplanes, planes * block.expansion, 1),
                norm_layer(planes * block.expansion),
            ]
            # print(downsample)
            downsample = nn.Sequential(*downsample)

        layers = []
        layers.append(
            block(
                self.inplanes,
                planes,
                stride,
                downsample,
                groups,
                norm_layer,
                filter_size=filter_size,
            )
        )
        self.inplanes = planes * block.expansion
        for _ in range(1, blocks):
            layers.append(
                block(
                    self.inplanes,
                    planes,
                    groups=groups,
                    norm_layer=norm_layer,
                    filter_size=filter_size,
                )
            )

        return nn.Sequential(*layers)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)

        x = self.avgpool(x)
        x = x.view(x.size(0), -1)
        if self.keep_fc:
            x = self.fc(x)

        return x



from typing import Any, Tuple

import numpy as np
from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.table import Table
from astropy.wcs import WCS


def flatten(
    f: fits.HDUList,
    x: float,
    y: float,
    size: float,
    hduid: int = 0,
    channel: int = 0,
    freqaxis: int = 3,
    verbose: bool = True,
) -> fits.HDUList:
    """
    Flatten a fits file so that it becomes a 2D image. Return new header and
    data
    This version also makes a sub-image of specified size.
    """

    naxis = f[hduid].header["NAXIS"]
    if naxis < 2:
        raise RuntimeError("Can't make map from this")

    if verbose:
        print(f[hduid].data.shape)
    data_shape = f[hduid].data.shape[-2:]
    by, bx = data_shape
    xmin = int(x - size / 2)
    if xmin < 0:
        xmin = 0
    xmax = int(x + size / 2)
    if xmax > bx:
        xmax = bx
    ymin = int(y - size / 2)
    if ymin < 0:
        ymin = 0
    ymax = int(y + size / 2)
    if ymax > by:
        ymax = by

    if ymax <= ymin or xmax <= xmin:
        # this can only happen if the required position is not on the map
        print(xmin, xmax, ymin, ymax)
        raise RuntimeError("Failed to make subimage! Required position not on the map.")

    w = WCS(f[hduid].header)
    wn = WCS(naxis=2)

    wn.wcs.crpix[0] = w.wcs.crpix[0] - xmin
    wn.wcs.crpix[1] = w.wcs.crpix[1] - ymin
    wn.wcs.cdelt = w.wcs.cdelt[0:2]
    try:
        wn.wcs.pc = w.wcs.pc[0:2, 0:2]
    except AttributeError:
        pass  # pc is not present
    wn.wcs.crval = w.wcs.crval[0:2]
    wn.wcs.ctype[0] = w.wcs.ctype[0]
    wn.wcs.ctype[1] = w.wcs.ctype[1]

    header = wn.to_header()
    header["NAXIS"] = 2

    slice = []
    for i in range(naxis, 0, -1):
        if i == 1:
            slice.append(np.s_[xmin:xmax])
        elif i == 2:
            slice.append(np.s_[ymin:ymax])
        elif i == freqaxis:
            slice.append(channel)
        else:
            slice.append(0)
    if verbose:
        print(slice)

    hdu = fits.PrimaryHDU(f[hduid].data[tuple(slice)], header)
    copy = ("EQUINOX", "EPOCH", "BMAJ", "BMIN", "BPA")
    for k in copy:
        r = f[hduid].header.get(k)
        if r:
            hdu.header[k] = r
    if "TAN" in hdu.header["CTYPE1"]:
        hdu.header["LATPOLE"] = f[hduid].header["CRVAL2"]
    hdulist = fits.HDUList([hdu])
    return hdulist


def extract_subimage(
    filename: str,
    ra: float,
    dec: float,
    size: float,
    hduid: int = 0,
    verbose: bool = True,
) -> fits.HDUList:
    if verbose:
        print("Opening", filename)
    orighdu = fits.open(filename)
    psize = int((size / orighdu[hduid].header["CDELT2"]))
    if verbose:
        print(f"Size in Pixels: {psize}")
        print(f"Size in Arcseconds: {size}")

    ndims = orighdu[hduid].header["NAXIS"]
    pvect = np.zeros((1, ndims))
    lwcs = WCS(orighdu[hduid].header)
    pvect[0][0] = ra
    pvect[0][1] = dec
    imc = lwcs.wcs_world2pix(pvect, 0)
    x = imc[0][0]
    y = imc[0][1]
    hdu = flatten(orighdu, x, y, psize, hduid=hduid, verbose=verbose)
    return hdu


def determine_visible_catalogue_source_and_separation(
    ra: float, dec: float, size: float, catalogue: Table, verbose=False
) -> Tuple[Any, Any, Any, SkyCoord, SkyCoord]:
    """
    Find the sources in the catalogue that are visible in the cutout, and returns a smaller catalogue for that
    :param ra: Radio RA
    :param dec: Radio DEC
    :param wcs: WCS of Radio FITS files
    :param size: Size of cutout in degrees
    :param catalogue: Pan-AllWISE catalogue
    :param l_objects: LOFAR Value Added Catalogue objects
    :return: Subcatalog of catalogue that only contains sources near the radio source in the cutout size, as well as
    SkyCoord of their world coordinates
    """
    try:
        ra_array = np.array(catalogue["RA"], dtype=float)
        dec_array = np.array(catalogue["DEC"], dtype=float)
    except:
        ra_array = np.array(catalogue["ID_ra"], dtype=float)
        dec_array = np.array(catalogue["ID_dec"], dtype=float)
    sky_coords = SkyCoord(ra_array, dec_array, unit="deg")

    source_coord = SkyCoord(ra, dec, unit="deg")
    search_radius = size * u.deg
    d2d = source_coord.separation(sky_coords)
    catalogmask = d2d < search_radius
    idxcatalog = np.where(catalogmask)[0]
    objects = catalogue[idxcatalog]

    try:
        ra_array = np.array(objects["RA"], dtype=float)
        dec_array = np.array(objects["DEC"], dtype=float)
    except:
        ra_array = np.array(objects["ID_ra"], dtype=float)
        dec_array = np.array(objects["ID_dec"], dtype=float)
    sky_coords = SkyCoord(ra_array, dec_array, unit="deg")
    d2d = source_coord.separation(sky_coords)
    angles = source_coord.position_angle(sky_coords)

    return objects, d2d, angles, source_coord, sky_coords


def determine_visible_catalogue_sources(
    ra: float, dec: float, size: float, catalogue: Table, verbose=False
) -> Table:
    """
    Find the sources in the catalogue that are visible in the cutout, and returns a smaller catalogue for that
    :param ra: Radio RA
    :param dec: Radio DEC
    :param wcs: WCS of Radio FITS files
    :param size: Size of cutout in degrees
    :param catalogue: Pan-AllWISE catalogue
    :param l_objects: LOFAR Value Added Catalogue objects
    :return: Subcatalog of catalogue that only contains sources near the radio source in the cutout size, as well as
    SkyCoord of their world coordinates
    """
    try:
        ra_array = np.array(catalogue["ra"], dtype=float)
        dec_array = np.array(catalogue["dec"], dtype=float)
    except:
        ra_array = np.array(catalogue["ID_ra"], dtype=float)
        dec_array = np.array(catalogue["ID_dec"], dtype=float)
    sky_coords = SkyCoord(ra_array, dec_array, unit="deg")

    source_coord = SkyCoord(ra, dec, unit="deg")
    search_radius = size * u.deg
    d2d = source_coord.separation(sky_coords)
    catalogmask = d2d < search_radius
    idxcatalog = np.where(catalogmask)[0]
    objects = catalogue[idxcatalog]

    return objects



from typing import Union, Optional, List, Tuple, Any

import imgaug as ia
import imgaug.augmenters as iaa
import numpy as np
from astropy.visualization import (
    MinMaxInterval,
    SqrtStretch,
    ManualInterval,
    ImageNormalize,
)
from imgaug.augmentables.bbs import BoundingBox, BoundingBoxesOnImage
from skimage.transform import rotate


def convert_to_valid_color(
    image_color: np.ndarray,
    clip: bool = False,
    lower_clip: float = 0.0,
    upper_clip: float = 1.0,
    normalize: bool = False,
    scaling: Optional[str] = None,
    simple_norm: bool = False,
) -> np.ndarray:
    """
    Convert the channel to a valid 0-1 range for RGB images
    """

    if simple_norm:
        interval = MinMaxInterval()
        norm = ImageNormalize()
        return norm(interval(image_color))

    if clip:
        image_color = np.clip(image_color, lower_clip, upper_clip)
    if normalize:
        interval = ManualInterval(lower_clip, upper_clip)
    else:
        interval = MinMaxInterval()
    if scaling == "sqrt":
        stretch = SqrtStretch()
        image_color = stretch(interval(image_color))
    else:
        norm = ImageNormalize()
        image_color = norm(interval(image_color))

    return image_color


def rotate_image_stacks(
    image_stack: np.ndarray, rotation_angle: Union[float, int] = 0
) -> np.ndarray:
    image_stack = rotate(
        image_stack,
        -rotation_angle,
        resize=True,
        center=None,
        order=1,
        mode="constant",
        cval=0,
        clip=True,
        preserve_range=True,
    )
    return image_stack


def make_bounding_box(
    source_location: Union[List[List[Union[float, int]]], np.ndarray]
) -> Tuple[
    int,
    int,
    int,
    int,
    Union[List[Union[float, int]], Any],
    Union[List[Union[float, int]], Any],
]:
    """
    Creates a bounding box and returns it in (xmin, ymin, xmax, ymax, class_name, source location) format
    for use with the source location in pixel values
    :return: Bounding box coordinates for COCO style annotation
    """
    # Now create box, which will be accomplished by taking int to get xmin, ymin, and int + 1 for xmax, ymax
    xmin = int(np.floor(source_location[5][0])) - 1
    ymin = int(np.floor(source_location[5][1])) - 1
    ymax = ymin + 2
    xmax = xmin + 2

    return xmin, ymin, xmax, ymax, source_location[4], source_location[5]


def augment_image_and_bboxes(
    image: np.ndarray,
    cutouts: Union[List[Tuple[float]], np.ndarray],
    proposal_boxes: Union[List[Tuple[float]], np.ndarray],
    angle: Optional[Union[float, int]],
    new_size: Optional[Union[int, Tuple[int]]],
    verbose: bool = False,
) -> Tuple[Any, Union[List[Tuple[float]], np.ndarray], np.ndarray]:
    bounding_boxes = []
    prop_boxes = []
    seg_boxes = []
    new_crop = int(
        np.ceil(image.shape[0] / np.sqrt(2))
    )  # Is 200 for 282, which is what is wanted, and works for others
    seq = iaa.Sequential(
        [
            iaa.Affine(rotate=angle),
            iaa.CropToFixedSize(width=new_crop, height=new_crop, position="center"),
        ]
    )
    for cutout in cutouts:
        bounding_boxes.append(
            BoundingBox(
                cutout[1] + 0.5, cutout[0] + 0.5, cutout[3] + 0.5, cutout[2] + 0.5
            )
        )
    for pbox in proposal_boxes:
        prop_boxes.append(
            BoundingBox(pbox[1] + 0.5, pbox[0] + 0.5, pbox[3] + 0.5, pbox[2] + 0.5)
        )

    bbs = BoundingBoxesOnImage(bounding_boxes, shape=image.shape)
    pbbs = BoundingBoxesOnImage(prop_boxes, shape=image.shape)
    if seg_boxes:
        sbbs = BoundingBoxesOnImage(seg_boxes, shape=image.shape)
        _, sbbs = seq(image=image, bounding_boxes=sbbs)
    _, bbs = seq(image=image, bounding_boxes=bbs)
    image, pbbs = seq(image=image, bounding_boxes=pbbs)
    # Rescale image and bounding boxes
    if type(new_size) == int:
        image_rescaled = ia.imresize_single_image(image, (new_size, new_size))
    else:
        image_rescaled = ia.imresize_single_image(
            image, (image.shape[0], image.shape[1])
        )
    bbs_rescaled = bbs.on(image_rescaled)
    pbbs_rescaled = pbbs.on(image_rescaled)
    if seg_boxes:
        sbbs_rescaled = sbbs.on(image_rescaled)
        sbbs_rescaled = sbbs_rescaled.remove_out_of_image(
            partly=False
        ).clip_out_of_image()

    # Remove bounding boxes that go out of bounds
    pbbs_rescaled = pbbs_rescaled.remove_out_of_image(partly=False).clip_out_of_image()
    # But only clip source bounding boxes that are partly out of frame, so that no sources are lost
    bbs_rescaled = bbs_rescaled.remove_out_of_image(partly=False).clip_out_of_image()
    # Also remove those and clip of segmentation maps
    for index, bbox in enumerate(bbs_rescaled):
        cutouts[index][0] = bbox.x1
        cutouts[index][1] = bbox.y1
        cutouts[index][2] = bbox.x2
        cutouts[index][3] = bbox.y2
    # Convert proposal boxes as well
    pbs = []
    for index, bbox in enumerate(pbbs_rescaled):
        pbs.append(np.asarray((bbox.x1, bbox.y1, bbox.x2, bbox.y2)))
    pbs = np.asarray(pbs)
    sbs = []
    try:
        for index, bbox in enumerate(sbbs_rescaled):
            sbs.append(np.asarray((bbox.x1, bbox.y1, bbox.x2, bbox.y2)))
    except:
        if verbose:
            print("Empty sbbs_rescaled")
    sbs = np.asarray(sbs)
    return image_rescaled, cutouts, pbs



import torch
import torch.nn as nn
import numpy as np
import torch.nn.functional as F


"""

Taken from https://github.com/pytorch/vision/blob/master/torchvision/models/resnet.py and modified to work with variable
input channels for working with grayscale radio data, and 10 channel multispectral data

    To create ResNet model would then be for grayscale
    kwargs['groups'] = 32
    kwargs['width_per_group'] = 8
    return ResNet(1, Bottleneck, [3, 4, 23, 3], **kwargs)
    
    And for 10-channel
    kwargs['groups'] = 32
    kwargs['width_per_group'] = 8
    return ResNet(10, Bottleneck, [3, 4, 23, 3], **kwargs)

"""


def conv3x3(in_planes, out_planes, stride=1, groups=1, dilation=1):
    """3x3 convolution with padding"""
    return nn.Conv2d(
        in_planes,
        out_planes,
        kernel_size=3,
        stride=stride,
        padding=dilation,
        groups=groups,
        bias=False,
        dilation=dilation,
    )


def conv1x1(in_planes, out_planes, stride=1):
    """1x1 convolution"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=1, stride=stride, bias=False)


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(
        self,
        inplanes,
        planes,
        stride=1,
        downsample=None,
        groups=1,
        base_width=64,
        dilation=1,
        norm_layer=None,
    ):
        super(BasicBlock, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if groups != 1 or base_width != 64:
            raise ValueError("BasicBlock only supports groups=1 and base_width=64")
        if dilation > 1:
            raise NotImplementedError("Dilation > 1 not supported in BasicBlock")
        # Both self.conv1 and self.downsample layers downsample the input when stride != 1
        self.conv1 = conv3x3(inplanes, planes, stride)
        self.bn1 = norm_layer(planes)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(planes, planes)
        self.bn2 = norm_layer(planes)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out


class Bottleneck(nn.Module):
    # Bottleneck in torchvision places the stride for downsampling at 3x3 convolution(self.conv2)
    # while original implementation places the stride at the first 1x1 convolution(self.conv1)
    # according to "Deep residual learning for image recognition"https://arxiv.org/abs/1512.03385.
    # This variant is also known as ResNet V1.5 and improves accuracy according to
    # https://ngc.nvidia.com/catalog/model-scripts/nvidia:resnet_50_v1_5_for_pytorch.

    expansion = 4

    def __init__(
        self,
        inplanes,
        planes,
        stride=1,
        downsample=None,
        groups=1,
        base_width=64,
        dilation=1,
        norm_layer=None,
    ):
        super(Bottleneck, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        width = int(planes * (base_width / 64.0)) * groups
        # Both self.conv2 and self.downsample layers downsample the input when stride != 1
        self.conv1 = conv1x1(inplanes, width)
        self.bn1 = norm_layer(width)
        self.conv2 = conv3x3(width, width, stride, groups, dilation)
        self.bn2 = norm_layer(width)
        self.conv3 = conv1x1(width, planes * self.expansion)
        self.bn3 = norm_layer(planes * self.expansion)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out


class ResNet(nn.Module):
    def __init__(
        self,
        num_channels,
        block,
        layers,
        num_classes=2,
        zero_init_residual=False,
        groups=1,
        width_per_group=64,
        replace_stride_with_dilation=None,
        norm_layer=None,
        keep_fc=True,
    ):
        super(ResNet, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        self._norm_layer = norm_layer
        self.keep_fc = keep_fc

        self.inplanes = 64
        self.dilation = 1
        if replace_stride_with_dilation is None:
            # each element in the tuple indicates if we should replace
            # the 2x2 stride with a dilated convolution instead
            replace_stride_with_dilation = [False, False, False]
        if len(replace_stride_with_dilation) != 3:
            raise ValueError(
                "replace_stride_with_dilation should be None "
                "or a 3-element tuple, got {}".format(replace_stride_with_dilation)
            )
        self.groups = groups
        self.base_width = width_per_group
        self.conv1 = nn.Conv2d(
            num_channels, self.inplanes, kernel_size=7, stride=2, padding=3, bias=False
        )
        self.bn1 = norm_layer(self.inplanes)
        self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
        self.layer1 = self._make_layer(block, 64, layers[0])
        self.layer2 = self._make_layer(
            block, 128, layers[1], stride=2, dilate=replace_stride_with_dilation[0]
        )
        self.layer3 = self._make_layer(
            block, 256, layers[2], stride=2, dilate=replace_stride_with_dilation[1]
        )
        self.layer4 = self._make_layer(
            block, 512, layers[3], stride=2, dilate=replace_stride_with_dilation[2]
        )
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = nn.Linear(512 * block.expansion, num_classes)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")
            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

        # Zero-initialize the last BN in each residual branch,
        # so that the residual branch starts with zeros, and each residual block behaves like an identity.
        # This improves the model by 0.2~0.3% according to https://arxiv.org/abs/1706.02677
        if zero_init_residual:
            for m in self.modules():
                if isinstance(m, Bottleneck):
                    nn.init.constant_(m.bn3.weight, 0)
                elif isinstance(m, BasicBlock):
                    nn.init.constant_(m.bn2.weight, 0)

    def _make_layer(self, block, planes, blocks, stride=1, dilate=False):
        norm_layer = self._norm_layer
        downsample = None
        previous_dilation = self.dilation
        if dilate:
            self.dilation *= stride
            stride = 1
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                conv1x1(self.inplanes, planes * block.expansion, stride),
                norm_layer(planes * block.expansion),
            )

        layers = []
        layers.append(
            block(
                self.inplanes,
                planes,
                stride,
                downsample,
                self.groups,
                self.base_width,
                previous_dilation,
                norm_layer,
            )
        )
        self.inplanes = planes * block.expansion
        for _ in range(1, blocks):
            layers.append(
                block(
                    self.inplanes,
                    planes,
                    groups=self.groups,
                    base_width=self.base_width,
                    dilation=self.dilation,
                    norm_layer=norm_layer,
                )
            )

        return nn.Sequential(*layers)

    def _forward_impl(self, x):
        # See note [TorchScript super()]
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)

        x = self.avgpool(x)
        x = torch.flatten(x, 1)
        if self.keep_fc:
            x = self.fc(x)

        return x

    def forward(self, x):
        return self._forward_impl(x)


class BinaryFocalLoss(nn.Module):
    """
    This is a implementation of Focal Loss with smooth label cross entropy supported which is proposed in
    'Focal Loss for Dense Object Detection. (https://arxiv.org/abs/1708.02002)'
        Focal_Loss= -1*alpha*(1-pt)*log(pt)
    :param num_class:
    :param alpha: (tensor) 3D or 4D the scalar factor for this criterion
    :param gamma: (float,double) gamma > 0 reduces the relative loss for well-classified examples (p>0.5) putting more
                    focus on hard misclassified example
    :param reduction: `none`|`mean`|`sum`
    :param **kwargs
        balance_index: (int) balance class index, should be specific when alpha is float
    """

    def __init__(self, alpha=[1.0, 1.0], gamma=2, ignore_index=None, reduction="mean"):
        super(BinaryFocalLoss, self).__init__()
        if alpha is None:
            alpha = [0.25, 0.75]
        self.alpha = alpha
        self.gamma = gamma
        self.smooth = 1e-6
        self.ignore_index = ignore_index
        self.reduction = reduction

        assert self.reduction in ["none", "mean", "sum"]

        if self.alpha is None:
            self.alpha = torch.ones(2)
        elif isinstance(self.alpha, (list, np.ndarray)):
            self.alpha = np.asarray(self.alpha)
            self.alpha = np.reshape(self.alpha, (2))
            assert (
                self.alpha.shape[0] == 2
            ), "the `alpha` shape is not match the number of class"
        elif isinstance(self.alpha, (float, int)):
            self.alpha = np.asarray(
                [self.alpha, 1.0 - self.alpha], dtype=np.float
            ).view(2)

        else:
            raise TypeError("{} not supported".format(type(self.alpha)))

    def forward(self, output, target):
        prob = torch.sigmoid(output)
        prob = torch.clamp(prob, self.smooth, 1.0 - self.smooth)

        pos_mask = (target == 1).float()
        neg_mask = (target == 0).float()

        pos_loss = (
            -self.alpha[0]
            * torch.pow(torch.sub(1.0, prob), self.gamma)
            * torch.log(prob)
            * pos_mask
        )
        neg_loss = (
            -self.alpha[1]
            * torch.pow(prob, self.gamma)
            * torch.log(torch.sub(1.0, prob))
            * neg_mask
        )

        neg_loss = neg_loss.sum()
        pos_loss = pos_loss.sum()
        num_pos = pos_mask.view(pos_mask.size(0), -1).sum()
        num_neg = neg_mask.view(neg_mask.size(0), -1).sum()

        if num_pos == 0:
            loss = neg_loss
        else:
            loss = pos_loss / num_pos + neg_loss / num_neg
        return loss



from detectron2.data import transforms as T
from detectron2.data import detection_utils as utils
import copy
import logging
import numpy as np
import torch

"""
This file contains the mapping that's applied to "dataset dicts" for LOFAR Source finding.
"""

__all__ = ["SourceMapper"]


class SourceMapper:
    """
    A callable which takes a dataset dict in Detectron2 Dataset format,
    and map it into a format used by the model.

    This is the default callable to be used to map your dataset dict into training data.
    You may need to follow it to implement your own one for customized logic,
    such as a different way to read or transform images.
    See :doc:`/tutorials/data_loading` for details.

    The callable currently does the following:

    1. Read the image from "file_name"
    2. Applies cropping/geometric transforms to the image and annotations
    3. Prepare data and annotations to Tensor and :class:`Instances`
    """

    def __init__(self, cfg, is_train=True):
        self.augmentation = utils.build_augmentation(cfg, is_train)
        if cfg.INPUT.CROP.ENABLED and is_train:
            self.augmentation.insert(
                0, T.RandomCrop(cfg.INPUT.CROP.TYPE, cfg.INPUT.CROP.SIZE)
            )
            logging.getLogger(__name__).info(
                "Cropping used in training: " + str(self.augmentation[0])
            )
            self.compute_tight_boxes = True
        else:
            self.compute_tight_boxes = False

        # fmt: off
        self.img_format     = cfg.INPUT.FORMAT
        self.mask_on        = cfg.MODEL.MASK_ON
        self.mask_format    = cfg.INPUT.MASK_FORMAT
        self.keypoint_on    = cfg.MODEL.KEYPOINT_ON
        self.load_proposals = cfg.MODEL.LOAD_PROPOSALS
        # fmt: on
        if self.keypoint_on and is_train:
            # Flip only makes sense in training
            self.keypoint_hflip_indices = utils.create_keypoint_hflip_indices(
                cfg.DATASETS.TRAIN
            )
        else:
            self.keypoint_hflip_indices = None

        if self.load_proposals:
            self.proposal_min_box_size = cfg.MODEL.PROPOSAL_GENERATOR.MIN_SIZE
            self.proposal_topk = (
                cfg.DATASETS.PRECOMPUTED_PROPOSAL_TOPK_TRAIN
                if is_train
                else cfg.DATASETS.PRECOMPUTED_PROPOSAL_TOPK_TEST
            )
        self.is_train = is_train

    def __call__(self, dataset_dict):
        """
        Args:
            dataset_dict (dict): Metadata of one image, in Detectron2 Dataset format.

        Returns:
            dict: a format that builtin models in detectron2 accept
        """
        dataset_dict = copy.deepcopy(dataset_dict)  # it will be modified by code below
        # USER: Write your own image loading if it's not from a file
        image = np.nan_to_num(np.load(dataset_dict["file_name"], allow_pickle=True))
        utils.check_image_size(dataset_dict, image)
        # USER: Remove if you don't do semantic/panoptic segmentation.
        if "sem_seg_file_name" in dataset_dict:
            sem_seg_gt = None  # utils.read_image(dataset_dict.pop("sem_seg_file_name"), "L").squeeze(2)
        else:
            sem_seg_gt = None

        aug_input = T.StandardAugInput(image, sem_seg=sem_seg_gt)
        transforms = aug_input.apply_augmentations(self.augmentation)
        image, sem_seg_gt = aug_input.image, aug_input.sem_seg

        image_shape = image.shape[:2]  # h, w
        # Pytorch's dataloader is efficient on torch.Tensor due to shared-memory,
        # but not efficient on large generic data structures due to the use of pickle & mp.Queue.
        # Therefore it's important to use torch.Tensor.
        dataset_dict["image"] = torch.as_tensor(
            np.ascontiguousarray(image.transpose(2, 0, 1)).astype("float32")
        )
        if sem_seg_gt is not None:
            dataset_dict["sem_seg"] = torch.as_tensor(sem_seg_gt.astype("long"))

        # USER: Remove if you don't use pre-computed proposals.
        # Most users would not need this feature.
        if self.load_proposals:
            utils.transform_proposals(
                dataset_dict,
                image_shape,
                transforms,
                proposal_topk=self.proposal_topk,
                min_box_size=self.proposal_min_box_size,
            )

        if not self.is_train:
            # USER: Modify this if you want to keep them for some reason.
            dataset_dict.pop("annotations", None)
            dataset_dict.pop("sem_seg_file_name", None)
            return dataset_dict

        if "annotations" in dataset_dict:
            # USER: Modify this if you want to keep them for some reason.
            for anno in dataset_dict["annotations"]:
                if not self.mask_on:
                    anno.pop("segmentation", None)
                if not self.keypoint_on:
                    anno.pop("keypoints", None)

            # USER: Implement additional transformations if you have other types of data
            annos = [
                utils.transform_instance_annotations(
                    obj,
                    transforms,
                    image_shape,
                    keypoint_hflip_indices=self.keypoint_hflip_indices,
                )
                for obj in dataset_dict.pop("annotations")
                if obj.get("iscrowd", 0) == 0
            ]
            instances = utils.annotations_to_instances(
                annos, image_shape, mask_format=self.mask_format
            )

            # After transforms such as cropping are applied, the bounding box may no longer
            # tightly bound the object. As an example, imagine a triangle object
            # [(0,0), (2,0), (0,2)] cropped by a box [(1,0),(2,2)] (XYXY format). The tight
            # bounding box of the cropped triangle should be [(1,0),(2,1)], which is not equal to
            # the intersection of original bounding box and the cropping box.
            if self.compute_tight_boxes and instances.has("gt_masks"):
                instances.gt_boxes = instances.gt_masks.get_bounding_boxes()
            dataset_dict["instances"] = utils.filter_empty_instances(instances)
        return dataset_dict



import pickle
from typing import Dict, Tuple, Any, Optional

import numpy as np
from astropy.io import fits
from astropy.table import Table


def get_lotss_objects(fname: str, verbose: bool = False) -> Table:
    """
    Load the LoTSS objects from a file
    """

    with fits.open(fname) as hdul:
        table = hdul[1].data
    print(table.columns)
    if verbose:
        print(table.columns)

    # convert from astropy.io.fits.fitsrec.FITS_rec to astropy.table.table.Table
    return Table(table)


def get_source_from_dict(source: Dict[str, Any]) -> str:
    if ".npy" in source["file_name"]:
        return source["file_name"].split("/")[-1].split(".npy")[0]
    else:
        return source["file_name"].split("/")[-1].split(".png")[0]


def get_lofar_dicts(annotation_filepath: str, fraction: float = 1.0) -> Dict[str, str]:
    with open(annotation_filepath, "rb") as f:
        dataset_dicts = pickle.load(f)
    if fraction < 0.99999:
        # Only take subset of the dataset
        num_entries = len(dataset_dicts)
        num_kept = int(fraction * num_entries)
        step_size = int(num_entries / num_kept)
        new_dicts = []
        for i in range(0, len(dataset_dicts), step_size):
            new_dicts.append(dataset_dicts[i])
        dataset_dicts = new_dicts
    return dataset_dicts


def get_only_mutli_dicts(
    annotation_filepath: str, multi: bool = True, vac: str = ".."
) -> list:
    """
    Only get the multi or single component sources, whichever is wanted
    """
    vac_catalog = get_lotss_objects(vac)
    if multi:
        vac_catalog = vac_catalog[vac_catalog["LGZ_Assoc"] > 1]
    else:
        vac_catalog = vac_catalog[vac_catalog["LGZ_Assoc"] == 1]
    vac_catalog = vac_catalog["Source_Name"].data
    with open(annotation_filepath, "rb") as f:
        dataset_dicts = pickle.load(f)
        new_dicts = []
        for i in range(0, len(dataset_dicts)):
            name = dataset_dicts[i]["file_name"].split("/")[-1].split(".npy")[0]
            if (
                name in vac_catalog or name.rpartition(".")[0] in vac_catalog
            ):  # Needed for both rotated and non rotated
                new_dicts.append(dataset_dicts[i])
        dataset_dicts = new_dicts
    return dataset_dicts


def make_physical_dict(
    vac_catalog: str,
    size_cut: float = 0.0,
    flux_cut: float = 0.0,
    multi: bool = False,
    lgz: bool = True,
) -> Dict[str, str]:
    """
    Makes cuts in the value added catalog for the size, flux, and multi vs single component
    :param vac_catalog: Value-added catalog location
    :param size_cut: The angular size, in arcseconds, to make a cut into large and small ones
    :param flux_cut: The low range to cut into high brightness vs low brightness
    :param multi: Whether to split on multi component or not
    :return:
    """
    vac_catalog = get_lotss_objects(vac_catalog)
    physical_cut = {}
    if lgz:
        vac_catalog = vac_catalog[~np.isnan(vac_catalog["LGZ_Size"])]
    if multi:
        all_multi = vac_catalog[vac_catalog["LGZ_Assoc"] > 1]
        all_single = vac_catalog[vac_catalog["LGZ_Assoc"] == 1]
        physical_cut["multi_comp"] = all_multi["Source_Name"].data
        physical_cut["single_comp"] = all_single["Source_Name"].data
        physical_cut[f"size{size_cut}"] = vac_catalog[
            vac_catalog["LGZ_Size"] >= size_cut
        ]["Source_Name"].data
        physical_cut[f"flux{flux_cut}"] = vac_catalog[
            vac_catalog["Total_flux"] >= flux_cut
        ]["Source_Name"].data
        physical_cut[f"fluxLess{flux_cut}"] = vac_catalog[
            vac_catalog["Total_flux"] < flux_cut
        ]["Source_Name"].data
        physical_cut[f"sizeLess{size_cut}"] = vac_catalog[
            vac_catalog["LGZ_Size"] < size_cut
        ]["Source_Name"].data
        physical_cut[f"size{size_cut}_flux{flux_cut}"] = vac_catalog[
            (vac_catalog["LGZ_Size"] >= size_cut)
            & (vac_catalog["Total_flux"] >= flux_cut)
        ]["Source_Name"].data
        physical_cut[f"sizeLess{size_cut}_flux{flux_cut}"] = vac_catalog[
            (vac_catalog["LGZ_Size"] < size_cut)
            & (vac_catalog["Total_flux"] >= flux_cut)
        ]["Source_Name"].data
        physical_cut[f"size{size_cut}_fluxLess{flux_cut}"] = vac_catalog[
            (vac_catalog["LGZ_Size"] >= size_cut)
            & (vac_catalog["Total_flux"] < flux_cut)
        ]["Source_Name"].data
        physical_cut[f"sizeLess{size_cut}_fluxLess{flux_cut}"] = vac_catalog[
            (vac_catalog["LGZ_Size"] < size_cut)
            & (vac_catalog["Total_flux"] < flux_cut)
        ]["Source_Name"].data
        physical_cut[f"multi_size{size_cut}_flux{flux_cut}"] = all_multi[
            (all_multi["LGZ_Size"] >= size_cut) & (all_multi["Total_flux"] >= flux_cut)
        ]["Source_Name"].data
        physical_cut[f"multi_sizeLess{size_cut}_flux{flux_cut}"] = all_multi[
            (all_multi["LGZ_Size"] < size_cut) & (all_multi["Total_flux"] >= flux_cut)
        ]["Source_Name"].data
        physical_cut[f"single_size{size_cut}_flux{flux_cut}"] = all_single[
            (all_single["LGZ_Size"] >= size_cut)
            & (all_single["Total_flux"] >= flux_cut)
        ]["Source_Name"].data
        physical_cut[f"single_sizeLess{size_cut}_flux{flux_cut}"] = all_single[
            (all_single["LGZ_Size"] < size_cut) & (all_single["Total_flux"] >= flux_cut)
        ]["Source_Name"].data
        physical_cut[f"multi_size{size_cut}_fluxLess{flux_cut}"] = all_multi[
            (all_multi["LGZ_Size"] >= size_cut) & (all_multi["Total_flux"] < flux_cut)
        ]["Source_Name"].data
        physical_cut[f"multi_sizeLess{size_cut}_fluxLess{flux_cut}"] = all_multi[
            (all_multi["LGZ_Size"] < size_cut) & (all_multi["Total_flux"] < flux_cut)
        ]["Source_Name"].data
        physical_cut[f"single_size{size_cut}_fluxLess{flux_cut}"] = all_single[
            (all_single["LGZ_Size"] >= size_cut) & (all_single["Total_flux"] < flux_cut)
        ]["Source_Name"].data
        physical_cut[f"single_sizeLess{size_cut}_fluxLess{flux_cut}"] = all_single[
            (all_single["LGZ_Size"] < size_cut) & (all_single["Total_flux"] < flux_cut)
        ]["Source_Name"].data
    else:
        physical_cut[f"size{size_cut}"] = vac_catalog[
            vac_catalog["LGZ_Size"] >= size_cut
        ]["Source_Name"].data
        physical_cut[f"flux{flux_cut}"] = vac_catalog[
            vac_catalog["Total_flux"] >= flux_cut
        ]["Source_Name"].data
        physical_cut[f"fluxLess{flux_cut}"] = vac_catalog[
            vac_catalog["Total_flux"] < flux_cut
        ]["Source_Name"].data
        physical_cut[f"sizeLess{size_cut}"] = vac_catalog[
            vac_catalog["LGZ_Size"] < size_cut
        ]["Source_Name"].data
        physical_cut[f"size{size_cut}_flux{flux_cut}"] = vac_catalog[
            (vac_catalog["LGZ_Size"] >= size_cut)
            & (vac_catalog["Total_flux"] >= flux_cut)
        ]["Source_Name"].data
        physical_cut[f"sizeLess{size_cut}_flux{flux_cut}"] = vac_catalog[
            (vac_catalog["LGZ_Size"] < size_cut)
            & (vac_catalog["Total_flux"] >= flux_cut)
        ]["Source_Name"].data
        physical_cut[f"size{size_cut}_fluxLess{flux_cut}"] = vac_catalog[
            (vac_catalog["LGZ_Size"] >= size_cut)
            & (vac_catalog["Total_flux"] < flux_cut)
        ]["Source_Name"].data
        physical_cut[f"sizeLess{size_cut}_fluxLess{flux_cut}"] = vac_catalog[
            (vac_catalog["LGZ_Size"] < size_cut)
            & (vac_catalog["Total_flux"] < flux_cut)
        ]["Source_Name"].data
    for key in physical_cut.keys():
        print(f"Cut: {key} | Num Elements: {len(physical_cut[key])}")
    return physical_cut


def get_physical_dicts(
    annotation_filepath: str, size: float = 0, vac_catalog: Optional[str] = None
) -> Tuple[list, list, list, list]:
    """
        Takes a subset of the annotation path based on criteria, such as multi component sources or image size, and
        returns four list[dict], split along multi- vs single component sources and size
    :param annotation_filepath:
    :param multi: Whether to take only multi component sources (True), only single component sources (False)
    :param size: Whether to make a cut on the size, in arcseconds, 0 no cut, keeps ones where the size is larger than the cut
    :param vac_catalog: LOFAR value added catalog
    :return: Multiple dicts of sources for Detectron2, with the applied cuts, useful to calculating the recall for various physical issues
    These dicts might be empty
    """
    with open(annotation_filepath, "rb") as f:
        dataset_dicts = pickle.load(f)
    vac_catalog = get_lotss_objects(vac_catalog)
    multi_large = []
    multi_small = []
    single_large = []
    single_small = []
    for element in dataset_dicts:
        source_name = get_source_from_dict(element)
        vac_entry = vac_catalog[vac_catalog["Source_Name"] == source_name]
        if vac_entry["LGZ_Assoc"] > 1:
            if vac_entry["LGZ_Size"] >= size:
                multi_large.append(element)
            elif vac_entry["LGZ_Size"] < size:
                multi_small.append(element)
        elif vac_entry["LGZ_Assoc"] < 2:
            if vac_entry["LGZ_Size"] >= size:
                single_large.append(element)
            elif vac_entry["LGZ_Size"] < size:
                single_small.append(element)
    return multi_large, multi_small, single_large, single_small



import os
from detectron2.data import (
    MetadataCatalog,
    build_detection_test_loader,
    build_detection_train_loader,
)

from lofarnn.models.dataloaders.SourceMapper import SourceMapper
from lofarnn.models.evaluators.SourceEvaluator import SourceEvaluator
from lofarnn.models.evaluators.LossEvalHook import LossEvalHook
from detectron2.engine import DefaultTrainer


class SourceTrainer(DefaultTrainer):
    """
    Modified DefaultTrainer to add in what's needed for LOFAR Source Detection

    Adds support for evaluating on multiple cuts in the data, given by physical_dict

    """

    _physical_dicts = None

    def __init__(self, cfg, physical_dict):
        super(SourceTrainer, self).__init__(cfg)
        SourceTrainer._physical_dicts = physical_dict

    @classmethod
    def build_evaluator(cls, cfg, dataset_name, output_folder=None, physical_dict=None):
        if output_folder is None:
            output_folder = os.path.join(cfg.OUTPUT_DIR, "inference")
        if physical_dict is None:
            physical_dict = cls._physical_dicts
        return SourceEvaluator(
            dataset_name, cfg, True, output_folder, physical_cut_dict=physical_dict
        )

    def build_hooks(self):
        hooks = super().build_hooks()
        hooks.insert(
            -1,
            LossEvalHook(
                self.cfg.TEST.EVAL_PERIOD,
                self.cfg.TEST.EXTRA_EVAL,
                self.model,
                build_detection_test_loader(
                    self.cfg, self.cfg.DATASETS.TEST[0], SourceMapper(self.cfg, True)
                ),
            ),
        )
        return hooks

    @classmethod
    def build_train_loader(cls, cfg):
        """
        Returns:
            iterable

        It now calls :func:`detectron2.data.build_detection_train_loader`.
        Overwrite it if you'd like a different data loader.
        """
        return build_detection_train_loader(cfg, mapper=SourceMapper(cfg))

    @classmethod
    def build_test_loader(cls, cfg, dataset_name):
        """
        Returns:
            iterable

        It now calls :func:`detectron2.data.build_detection_test_loader`.
        Overwrite it if you'd like a different data loader.
        """
        return build_detection_test_loader(
            cfg, dataset_name, mapper=SourceMapper(cfg, False)
        )



from typing import Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
import torch
import torch.nn.functional as F
from captum.attr import (
    Occlusion,
    Saliency,
    IntegratedGradients,
    NoiseTunnel,
    visualization,
    ShapleyValueSampling,
)

"""

Set of visualizations from a model to determine why those models look at what they do

"""


def score_func(o: torch.Tensor) -> torch.Tensor:
    return F.softmax(o, dim=1)


def convert_to_image(
    inputs: Tuple[torch.Tensor, torch.Tensor]
) -> Tuple[np.ndarray, np.ndarray]:
    return (
        np.transpose(inputs[0].cpu().detach().numpy(), (0, 2, 3, 1)),
        inputs[1].cpu().detach().numpy().reshape((1, 12, 1)),
    )


def convert_to_image_multi(
    inputs: Tuple[torch.Tensor, torch.Tensor]
) -> Tuple[np.ndarray, np.ndarray]:
    return (
        np.transpose(inputs[0].cpu().detach().numpy(), (0, 2, 3, 1)),
        np.transpose(inputs[1].cpu().detach().numpy(), (0, 2, 3, 1))[0],
    )


def visualize_maps(
    model: torch.nn.Module,
    inputs: Union[Tuple[torch.Tensor, torch.Tensor]],
    labels: torch.Tensor,
    title: str,
    second_occlusion: Tuple[int, int, int] = (1, 2, 2),
    baselines: Tuple[int, int] = (0, 0),
    closest: bool = False,
) -> None:
    """
    Visualizes the average of the inputs, or the single input, using various different XAI approaches
    """
    single = inputs[1].ndim == 2
    model.zero_grad()
    model.eval()
    occ = Occlusion(model)
    saliency = Saliency(model)
    saliency = NoiseTunnel(saliency)
    igrad = IntegratedGradients(model)
    igrad_2 = NoiseTunnel(igrad)
    # deep_lift = DeepLift(model)
    grad_shap = ShapleyValueSampling(model)
    output = model(inputs[0], inputs[1])
    output = F.softmax(output, dim=-1).argmax(dim=1, keepdim=True)
    labels = F.softmax(labels, dim=-1).argmax(dim=1, keepdim=True)
    if np.all(labels.cpu().numpy() == 1) and not closest:
        return
    if True:
        targets = labels
    else:
        targets = output
    print(targets)
    correct = targets.cpu().numpy() == labels.cpu().numpy()
    # if correct:
    #   return
    occ_out = occ.attribute(
        inputs,
        baselines=baselines,
        sliding_window_shapes=((1, 5, 5), second_occlusion),
        target=targets,
    )
    # occ_out2 = occ.attribute(inputs, sliding_window_shapes=((1,20,20), second_occlusion), strides=(8,1), target=targets)
    saliency_out = saliency.attribute(
        inputs, nt_type="smoothgrad_sq", n_samples=5, target=targets, abs=False
    )
    # igrad_out = igrad.attribute(inputs, target=targets, internal_batch_size=1)
    igrad_out = igrad_2.attribute(
        inputs,
        baselines=baselines,
        target=targets,
        n_samples=5,
        nt_type="smoothgrad_sq",
        internal_batch_size=1,
    )
    # deep_lift_out = deep_lift.attribute(inputs, target=targets)
    grad_shap_out = grad_shap.attribute(inputs, baselines=baselines, target=targets)

    if single:
        inputs = convert_to_image(inputs)
        occ_out = convert_to_image(occ_out)
        saliency_out = convert_to_image(saliency_out)
        igrad_out = convert_to_image(igrad_out)
        # grad_shap_out = convert_to_image(grad_shap_out)
    else:
        inputs = convert_to_image_multi(inputs)
        occ_out = convert_to_image_multi(occ_out)
        saliency_out = convert_to_image_multi(saliency_out)
        igrad_out = convert_to_image_multi(igrad_out)
        grad_shap_out = convert_to_image_multi(grad_shap_out)
    fig, axes = plt.subplots(2, 5)
    (fig, axes[0, 0]) = visualization.visualize_image_attr(
        occ_out[0][0],
        inputs[0][0],
        title="Original Image",
        method="original_image",
        show_colorbar=True,
        plt_fig_axis=(fig, axes[0, 0]),
        use_pyplot=False,
    )
    (fig, axes[0, 1]) = visualization.visualize_image_attr(
        occ_out[0][0],
        None,
        sign="all",
        title="Occ (5x5)",
        show_colorbar=True,
        plt_fig_axis=(fig, axes[0, 1]),
        use_pyplot=False,
    )
    (fig, axes[0, 2]) = visualization.visualize_image_attr(
        saliency_out[0][0],
        None,
        sign="all",
        title="Saliency",
        show_colorbar=True,
        plt_fig_axis=(fig, axes[0, 2]),
        use_pyplot=False,
    )
    (fig, axes[0, 3]) = visualization.visualize_image_attr(
        igrad_out[0][0],
        None,
        sign="all",
        title="Integrated Grad",
        show_colorbar=True,
        plt_fig_axis=(fig, axes[0, 3]),
        use_pyplot=False,
    )
    (fig, axes[0, 4]) = visualization.visualize_image_attr(
        grad_shap_out[0],
        None,
        title="GradSHAP",
        show_colorbar=True,
        plt_fig_axis=(fig, axes[0, 4]),
        use_pyplot=False,
    )
    ##### Second Input Labels #########################################################################################
    (fig, axes[1, 0]) = visualization.visualize_image_attr(
        occ_out[1],
        inputs[1],
        title="Original Aux",
        method="original_image",
        show_colorbar=True,
        plt_fig_axis=(fig, axes[1, 0]),
        use_pyplot=False,
    )
    (fig, axes[1, 1]) = visualization.visualize_image_attr(
        occ_out[1],
        None,
        sign="all",
        title="Occ (1x1)",
        show_colorbar=True,
        plt_fig_axis=(fig, axes[1, 1]),
        use_pyplot=False,
    )
    (fig, axes[1, 2]) = visualization.visualize_image_attr(
        saliency_out[1],
        None,
        sign="all",
        title="Saliency",
        show_colorbar=True,
        plt_fig_axis=(fig, axes[1, 2]),
        use_pyplot=False,
    )
    (fig, axes[1, 3]) = visualization.visualize_image_attr(
        igrad_out[1],
        None,
        sign="all",
        title="Integrated Grad",
        show_colorbar=True,
        plt_fig_axis=(fig, axes[1, 3]),
        use_pyplot=False,
    )
    (fig, axes[1, 4]) = visualization.visualize_image_attr(
        grad_shap_out[1],
        None,
        title="GradSHAP",
        show_colorbar=True,
        plt_fig_axis=(fig, axes[1, 4]),
        use_pyplot=False,
    )

    fig.suptitle(
        title + f" Label: {labels.cpu().numpy()} Pred: {targets.cpu().numpy()}"
    )
    plt.savefig(
        f"{title}_{'single' if single else 'multi'}_{'Failed' if correct else 'Success'}_baseline{baselines[0]}.png",
        dpi=300,
    )
    plt.clf()
    plt.cla()



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
import contextlib
import copy
import io
import itertools
import json
import logging
import numpy as np
import os
import pickle
from collections import OrderedDict
import pycocotools.mask as mask_util
import torch
from fvcore.common.file_io import PathManager
from pycocotools.coco import COCO
from pycocotools.cocoeval import COCOeval
from tabulate import tabulate

import detectron2.utils.comm as comm
from detectron2.data import MetadataCatalog
from detectron2.data.datasets.coco import convert_to_coco_json
from detectron2.structures import Boxes, BoxMode, pairwise_iou
from detectron2.utils.logger import create_small_table

from detectron2.evaluation.evaluator import DatasetEvaluator


class SourceEvaluator(DatasetEvaluator):
    """
    Evaluate object proposal, instance detection/segmentation, keypoint detection
    outputs using COCO's metrics and APIs. But only do the bboxes with the one with the single highest confidence

    Calculates it over multi-component, single-component, and size-limited subsets

    """

    def __init__(
        self, dataset_name, cfg, distributed, output_dir=None, physical_cut_dict=None
    ):
        """
        Args:
            dataset_name (str): name of the dataset to be evaluated.
                It must have either the following corresponding metadata:

                    "json_file": the path to the COCO format annotation

                Or it must be in detectron2's standard dataset format
                so it can be converted to COCO format automatically.
            cfg (CfgNode): config instance
            distributed (True): if True, will collect results from all ranks and run evaluation
                in the main process.
                Otherwise, will evaluate the results in the current process.
            output_dir (str): optional, an output directory to dump all
                results predicted on the dataset. The dump contains two files:

                1. "instance_predictions.pth" a file in torch serialization
                   format that contains all the raw original predictions.
                2. "coco_instances_results.json" a json file in COCO's result
                   format.
            physical_cut_dict (dict): Optional, Dictionary mapping Source Names to multicomponent,
             single component, and size cuts, should be in format of {cut: [list]} where list is the Source Names for that cut
        """
        self._tasks = self._tasks_from_config(cfg)
        self._distributed = distributed
        self._output_dir = output_dir
        self._physical_cuts = physical_cut_dict

        self._cpu_device = torch.device("cpu")
        self._logger = logging.getLogger(__name__)
        self._dataset_name = dataset_name

        self._metadata = MetadataCatalog.get(dataset_name)
        if not hasattr(self._metadata, "json_file"):
            self._logger.warning(
                f"json_file was not found in MetaDataCatalog for '{dataset_name}'."
                " Trying to convert it to COCO format ..."
            )

            cache_path = os.path.join(output_dir, f"{dataset_name}_coco_format.json")
            self._metadata.json_file = cache_path
            convert_to_coco_json(dataset_name, cache_path)

        json_file = PathManager.get_local_path(self._metadata.json_file)
        with contextlib.redirect_stdout(io.StringIO()):
            self._coco_api = COCO(json_file)

        self._kpt_oks_sigmas = cfg.TEST.KEYPOINT_OKS_SIGMAS
        # Test set json files do not contain annotations (evaluation must be
        # performed using the COCO evaluation server).
        self._do_evaluation = "annotations" in self._coco_api.dataset

    def reset(self):
        self._predictions = []

    def _tasks_from_config(self, cfg):
        """
        Returns:
            tuple[str]: tasks that can be evaluated under the given configuration.
        """
        tasks = ("bbox",)
        if cfg.MODEL.MASK_ON:
            tasks = tasks + ("segm",)
        if cfg.MODEL.KEYPOINT_ON:
            tasks = tasks + ("keypoints",)
        return tasks

    def process(self, inputs, outputs):
        """
        Args:
            inputs: the inputs to a COCO model (e.g., GeneralizedRCNN).
                It is a list of dict. Each dict corresponds to an image and
                contains keys like "height", "width", "file_name", "image_id".
            outputs: the outputs of a COCO model. It is a list of dicts with key
                "instances" that contains :class:`Instances`.
        """
        for input, output in zip(inputs, outputs):
            if ".npy" in input["file_name"]:
                source_name = input["file_name"].split("/")[-1].split(".npy")[0]
            else:
                source_name = input["file_name"].split("/")[-1].split(".png")[0]
            prediction = {
                "image_id": input["image_id"],
                "source_name": source_name,
            }  # Add Name to get component later

            # TODO this is ugly
            if "instances" in output:
                instances = output["instances"].to(self._cpu_device)
                prediction["instances"] = instances_to_coco_json(
                    instances, input["image_id"]
                )
            if "proposals" in output:
                prediction["proposals"] = output["proposals"].to(self._cpu_device)
            self._predictions.append(prediction)

    def evaluate(self):
        if self._distributed:
            comm.synchronize()
            predictions = comm.gather(self._predictions, dst=0)
            predictions = list(itertools.chain(*predictions))

            if not comm.is_main_process():
                return {}
        else:
            predictions = self._predictions

        if len(predictions) == 0:
            self._logger.warning("[COCOEvaluator] Did not receive valid predictions.")
            return {}

        if self._output_dir:
            PathManager.mkdirs(self._output_dir)
            file_path = os.path.join(self._output_dir, "instances_predictions.pth")
            with PathManager.open(file_path, "wb") as f:
                torch.save(predictions, f)

        self._results = OrderedDict()
        if "proposals" in predictions[0]:
            self._eval_box_proposals(predictions)
        if "instances" in predictions[0]:
            self._eval_predictions(set(self._tasks), predictions)
        # Copy so the caller can do whatever with results
        return copy.deepcopy(self._results)

    def _eval_predictions(self, tasks, predictions):
        """
        Evaluate predictions on the given tasks.
        Fill self._results with the metrics of the tasks.
        """
        self._logger.info("Preparing results for COCO format ...")
        coco_results = list(itertools.chain(*[x["instances"] for x in predictions]))

        # unmap the category ids for COCO
        if hasattr(self._metadata, "thing_dataset_id_to_contiguous_id"):
            reverse_id_mapping = {
                v: k
                for k, v in self._metadata.thing_dataset_id_to_contiguous_id.items()
            }
            for result in coco_results:
                category_id = result["category_id"]
                assert (
                    category_id in reverse_id_mapping
                ), "A prediction has category_id={}, which is not available in the dataset.".format(
                    category_id
                )
                result["category_id"] = reverse_id_mapping[category_id]

        if self._output_dir:
            file_path = os.path.join(self._output_dir, "coco_instances_results.json")
            self._logger.info("Saving results to {}".format(file_path))
            with PathManager.open(file_path, "w") as f:
                f.write(json.dumps(coco_results))
                f.flush()

        if not self._do_evaluation:
            self._logger.info("Annotations are not available for evaluation.")
            return

        # Calculate the recall based on general recall and precision, not COCO mAP, with single best prediction
        self._logger.info(f"Evaluating with non-mAR...")
        all_recall = _evaluate_box_proposals(predictions, self._coco_api, limit=1)
        pickle.dump(
            all_recall["per_source"],
            file=open(
                os.path.join(
                    self._output_dir, f"{self._dataset_name}_recall_limit1.pkl"
                ),
                "wb",
            ),
        )
        self._results["own_recall"] = {
            "ar": all_recall["ar"],
            "ap": all_recall["ap"],
            "precision": all_recall["precisions"][-1],
            "recall": all_recall["recalls"][-1],
        }
        all_recall = _evaluate_box_proposals(predictions, self._coco_api, limit=2)
        pickle.dump(
            all_recall["per_source"],
            file=open(
                os.path.join(
                    self._output_dir, f"{self._dataset_name}_recall_limit2.pkl"
                ),
                "wb",
            ),
        )
        self._results["own_recall_2"] = {
            "ar": all_recall["ar"],
            "ap": all_recall["ap"],
            "precision": all_recall["precisions"][-1],
            "recall": all_recall["recalls"][-1],
        }
        all_recall = _evaluate_box_proposals(predictions, self._coco_api, limit=5)
        pickle.dump(
            all_recall["per_source"],
            file=open(
                os.path.join(
                    self._output_dir, f"{self._dataset_name}_recall_limit5.pkl"
                ),
                "wb",
            ),
        )
        self._results["own_recall_5"] = {
            "ar": all_recall["ar"],
            "ap": all_recall["ap"],
            "precision": all_recall["precisions"][-1],
            "recall": all_recall["recalls"][-1],
        }
        all_recall = _evaluate_box_proposals(predictions, self._coco_api, limit=10)
        pickle.dump(
            all_recall["per_source"],
            file=open(
                os.path.join(
                    self._output_dir, f"{self._dataset_name}_recall_limit10.pkl"
                ),
                "wb",
            ),
        )
        self._results["own_recall_10"] = {
            "ar": all_recall["ar"],
            "ap": all_recall["ap"],
            "precision": all_recall["precisions"][-1],
            "recall": all_recall["recalls"][-1],
        }
        all_recall = _evaluate_box_proposals(predictions, self._coco_api, limit=100)
        pickle.dump(
            all_recall["per_source"],
            file=open(
                os.path.join(
                    self._output_dir, f"{self._dataset_name}_recall_limit100.pkl"
                ),
                "wb",
            ),
        )
        self._results["own_recall_100"] = {
            "ar": all_recall["ar"],
            "ap": all_recall["ap"],
            "precision": all_recall["precisions"][-1],
            "recall": all_recall["recalls"][-1],
        }
        for physical_cut in self._physical_cuts.keys():
            self._logger.info(
                f"Evaluating with non-mAR on physical cut {physical_cut}..."
            )
            prediction_cut = self._get_physical_cut_predictions(
                physical_cut, predictions
            )
            # physical_coco_results = list(itertools.chain(*[x["instances"] for x in prediction_cut]))
            phys_recall = _evaluate_box_proposals(
                prediction_cut, self._coco_api, limit=1
            )
            recall_name = f"own_recall_{physical_cut}"
            self._results[recall_name] = {
                "ar": phys_recall["ar"],
                "ap": phys_recall["ap"],
                "precision": phys_recall["precisions"][-1],
                "recall": phys_recall["recalls"][-1],
            }
            phys_recall = _evaluate_box_proposals(
                prediction_cut, self._coco_api, limit=2
            )
            recall_name = f"own_recall_2_{physical_cut}"
            self._results[recall_name] = {
                "ar": phys_recall["ar"],
                "ap": phys_recall["ap"],
                "precision": phys_recall["precisions"][-1],
                "recall": phys_recall["recalls"][-1],
            }
            phys_recall = _evaluate_box_proposals(
                prediction_cut, self._coco_api, limit=5
            )
            recall_name = f"own_recall_5_{physical_cut}"
            self._results[recall_name] = {
                "ar": phys_recall["ar"],
                "ap": phys_recall["ap"],
                "precision": phys_recall["precisions"][-1],
                "recall": phys_recall["recalls"][-1],
            }
            phys_recall = _evaluate_box_proposals(
                prediction_cut, self._coco_api, limit=10
            )
            recall_name = f"own_recall_10_{physical_cut}"
            self._results[recall_name] = {
                "ar": phys_recall["ar"],
                "ap": phys_recall["ap"],
                "precision": phys_recall["precisions"][-1],
                "recall": phys_recall["recalls"][-1],
            }
            phys_recall = _evaluate_box_proposals(
                prediction_cut, self._coco_api, limit=100
            )
            recall_name = f"own_recall_100_{physical_cut}"
            self._results[recall_name] = {
                "ar": phys_recall["ar"],
                "ap": phys_recall["ap"],
                "precision": phys_recall["precisions"][-1],
                "recall": phys_recall["recalls"][-1],
            }
        self._logger.info("Evaluating predictions ...")
        for task in sorted(tasks):
            coco_eval = (
                _evaluate_predictions_on_coco(
                    self._coco_api,
                    coco_results,
                    task,
                    kpt_oks_sigmas=self._kpt_oks_sigmas,
                )
                if len(coco_results) > 0
                else None  # cocoapi does not handle empty results very well
            )
            res = self._derive_coco_results(
                coco_eval, task, class_names=self._metadata.get("thing_classes")
            )
            self._results[task] = res
        for physical_cut in self._physical_cuts.keys():
            self._logger.info(f"Evaluating on physical cut {physical_cut}...")
            prediction_cut = self._get_physical_cut_predictions(
                physical_cut, predictions
            )
            physical_coco_results = list(
                itertools.chain(*[x["instances"] for x in prediction_cut])
            )
            for task in sorted(tasks):
                coco_eval = (
                    _evaluate_predictions_on_coco(
                        self._coco_api,
                        physical_coco_results,
                        task,
                        kpt_oks_sigmas=self._kpt_oks_sigmas,
                    )
                    if len(physical_coco_results) > 0
                    else None  # cocoapi does not handle empty results very well
                )
                res = self._derive_coco_results(
                    coco_eval, task, class_names=self._metadata.get("thing_classes")
                )
                result_name = f"{task}_{physical_cut}"
                self._results[result_name] = res

    def _get_physical_cut_predictions(self, cut_key, predictions):
        prediction_cut = []
        for prediction in predictions:
            if prediction["source_name"] in self._physical_cuts[cut_key]:
                prediction_cut.append(prediction)
            elif (
                prediction["source_name"].rpartition(".")[0]
                in self._physical_cuts[cut_key]
            ):
                # Handles the rotation, where there is an extra '.rot' after the source name
                prediction_cut.append(prediction)
        return prediction_cut

    def _derive_coco_results(self, coco_eval, iou_type, class_names=None):
        """
        Derive the desired score numbers from summarized COCOeval.

        Args:
            coco_eval (None or COCOEval): None represents no predictions from model.
            iou_type (str):
            class_names (None or list[str]): if provided, will use it to predict
                per-category AP.

        Returns:
            a dict of {metric name: score}
        """

        metrics = {
            "bbox": [
                "AP",
                "AP50",
                "AP75",
                "APs",
                "APm",
                "APl",
                "AR",
                "AR50",
                "AR75",
                "ARs",
                "ARm",
                "ARl",
            ],
            "segm": [
                "AP",
                "AP50",
                "AP75",
                "APs",
                "APm",
                "APl",
                "AR",
                "AR50",
                "AR75",
                "ARs",
                "ARm",
                "ARl",
            ],
            "keypoints": ["AP", "AP50", "AP75", "APm", "APl"],
        }[iou_type]

        if coco_eval is None:
            self._logger.warn("No predictions from the model!")
            return {metric: float("nan") for metric in metrics}

        # the standard metrics
        results = {
            metric: float(
                coco_eval.stats[idx] * 100 if coco_eval.stats[idx] >= 0 else "nan"
            )
            for idx, metric in enumerate(metrics)
        }
        self._logger.info(
            "Evaluation results for {}: \n".format(iou_type)
            + create_small_table(results)
        )
        if not np.isfinite(sum(results.values())):
            self._logger.info("Note that some metrics cannot be computed.")

        if class_names is None or len(class_names) <= 1:
            return results
        # Compute per-category AP
        # from https://github.com/facebookresearch/Detectron/blob/a6a835f5b8208c45d0dce217ce9bbda915f44df7/detectron/datasets/json_dataset_evaluator.py#L222-L252 # noqa
        precisions = coco_eval.eval["recall"]
        # precision has dims (iou, recall, cls, area range, max dets)
        assert len(class_names) == precisions.shape[2]

        results_per_category = []
        for idx, name in enumerate(class_names):
            # area range index 0: all area ranges
            # max dets index -1: typically 100 per image
            precision = precisions[:, :, idx, 0, -1]
            precision = precision[precision > -1]
            ap = np.mean(precision) if precision.size else float("nan")
            results_per_category.append(("{}".format(name), float(ap * 100)))

        # tabulate it
        N_COLS = min(6, len(results_per_category) * 2)
        results_flatten = list(itertools.chain(*results_per_category))
        results_2d = itertools.zip_longest(
            *[results_flatten[i::N_COLS] for i in range(N_COLS)]
        )
        table = tabulate(
            results_2d,
            tablefmt="pipe",
            floatfmt=".3f",
            headers=["category", "AP"] * (N_COLS // 2),
            numalign="left",
        )
        self._logger.info("Per-category {} AP: \n".format(iou_type) + table)

        results.update({"AP-" + name: ap for name, ap in results_per_category})
        return results


def instances_to_coco_json(instances, img_id):
    """
    Dump an "Instances" object to a COCO-format json that's used for evaluation.

    Args:
        instances (Instances):
        img_id (int): the image id

    Returns:
        list[dict]: list of json annotations in COCO format.
    """
    num_instance = len(instances)
    if num_instance == 0:
        return []

    boxes = instances.pred_boxes.tensor.numpy()
    boxes = BoxMode.convert(boxes, BoxMode.XYXY_ABS, BoxMode.XYWH_ABS)
    boxes = boxes.tolist()
    scores = instances.scores.tolist()
    classes = instances.pred_classes.tolist()

    has_mask = instances.has("pred_masks")
    if has_mask:
        # use RLE to encode the masks, because they are too large and takes memory
        # since this evaluator stores outputs of the entire dataset
        rles = [
            mask_util.encode(np.array(mask[:, :, None], order="F", dtype="uint8"))[0]
            for mask in instances.pred_masks
        ]
        for rle in rles:
            # "counts" is an array encoded by mask_util as a byte-stream. Python3's
            # json writer which always produces strings cannot serialize a bytestream
            # unless you decode it. Thankfully, utf-8 works out (which is also what
            # the pycocotools/_mask.pyx does).
            rle["counts"] = rle["counts"].decode("utf-8")

    has_keypoints = instances.has("pred_keypoints")
    if has_keypoints:
        keypoints = instances.pred_keypoints

    results = []
    for k in range(num_instance):
        result = {
            "image_id": img_id,
            "category_id": classes[k],
            "bbox": boxes[k],
            "score": scores[k],
        }
        if has_mask:
            result["segmentation"] = rles[k]
        if has_keypoints:
            # In COCO annotations,
            # keypoints coordinates are pixel indices.
            # However our predictions are floating point coordinates.
            # Therefore we subtract 0.5 to be consistent with the annotation format.
            # This is the inverse of data loading logic in `datasets/coco.py`.
            keypoints[k][:, :2] -= 0.5
            result["keypoints"] = keypoints[k].flatten().tolist()
        results.append(result)
    return results


# inspired from Detectron:
# https://github.com/facebookresearch/Detectron/blob/a6a835f5b8208c45d0dce217ce9bbda915f44df7/detectron/datasets/json_dataset_evaluator.py#L255 # noqa
def _evaluate_box_proposals(
    dataset_predictions, coco_api, thresholds=None, area="all", limit=None
):
    """
    Evaluate detection proposal recall metrics. This function is a much
    faster alternative to the official COCO API recall evaluation code. However,
    it produces slightly different results.
    """
    # Record max overlap value for each gt box
    # Return vector of overlap values
    areas = {
        "all": 0,
        "small": 1,
        "medium": 2,
        "large": 3,
        "96-128": 4,
        "128-256": 5,
        "256-512": 6,
        "512-inf": 7,
    }
    area_ranges = [
        [0 ** 2, 1e5 ** 2],  # all
        [0 ** 2, 32 ** 2],  # small
        [32 ** 2, 96 ** 2],  # medium
        [96 ** 2, 1e5 ** 2],  # large
        [96 ** 2, 128 ** 2],  # 96-128
        [128 ** 2, 256 ** 2],  # 128-256
        [256 ** 2, 512 ** 2],  # 256-512
        [512 ** 2, 1e5 ** 2],
    ]  # 512-inf
    assert area in areas, "Unknown area range: {}".format(area)
    area_range = area_ranges[areas[area]]
    gt_overlaps = []
    num_pos = 0
    num_boxes = 0
    source_outcomes = {}
    for prediction_dict in dataset_predictions:
        preds = np.asarray(prediction_dict["instances"])
        scores = []
        for i in preds:
            scores.append(i["score"])

        # sort predictions in descending order
        # TODO maybe remove this and make it explicit in the documentation
        inds = (-1 * np.asarray(scores)).argsort()
        preds = preds[inds]

        ann_ids = coco_api.getAnnIds(imgIds=prediction_dict["image_id"])
        anno = coco_api.loadAnns(ann_ids)
        gt_boxes = [
            BoxMode.convert(obj["bbox"], BoxMode.XYWH_ABS, BoxMode.XYXY_ABS)
            for obj in anno
            if obj["iscrowd"] == 0
        ]
        gt_boxes = torch.as_tensor(gt_boxes).reshape(-1, 4)  # guard against no boxes
        gt_boxes = Boxes(gt_boxes)
        gt_areas = torch.as_tensor([obj["area"] for obj in anno if obj["iscrowd"] == 0])

        if len(gt_boxes) == 0 or len(preds) == 0:
            continue

        valid_gt_inds = (gt_areas >= area_range[0]) & (gt_areas <= area_range[1])
        gt_boxes = gt_boxes[valid_gt_inds]

        num_pos += len(gt_boxes)  # 1 GT from each image
        num_boxes += limit  # N number of proposals kept

        if len(gt_boxes) == 0:
            continue

        if limit is not None and len(preds) > limit:
            preds = preds[:limit]

        pred_boxes = []
        for i in preds:
            pred_boxes.append(
                BoxMode.convert(i["bbox"], BoxMode.XYWH_ABS, BoxMode.XYXY_ABS)
            )
        pred_boxes = torch.as_tensor(pred_boxes).reshape(
            -1, 4
        )  # guard against no boxes
        pred_boxes = Boxes(pred_boxes)
        overlaps = pairwise_iou(pred_boxes, gt_boxes)

        _gt_overlaps = torch.zeros(len(gt_boxes))
        for j in range(min(len(preds), len(gt_boxes))):
            # find which proposal box maximally covers each gt box
            # and get the iou amount of coverage for each gt box
            max_overlaps, argmax_overlaps = overlaps.max(dim=0)

            # find which gt box is 'best' covered (i.e. 'best' = most iou)
            gt_ovr, gt_ind = max_overlaps.max(dim=0)
            assert gt_ovr >= 0
            # find the proposal box that covers the best covered gt box
            box_ind = argmax_overlaps[gt_ind]
            # record the iou coverage of this gt box
            _gt_overlaps[j] = overlaps[box_ind, gt_ind]
            assert _gt_overlaps[j] == gt_ovr
            # mark the proposal box and the gt box as used
            overlaps[box_ind, :] = -1
            overlaps[:, gt_ind] = -1
            print(preds[gt_ind])
            source_outcomes[
                prediction_dict["source_name"]
            ] = gt_ovr.item()  # Save overlap, so can be used for determining outcome

        # append recorded iou coverage level
        gt_overlaps.append(_gt_overlaps)
    gt_overlaps = (
        torch.cat(gt_overlaps, dim=0)
        if len(gt_overlaps)
        else torch.zeros(0, dtype=torch.float32)
    )
    gt_overlaps, _ = torch.sort(gt_overlaps)

    if thresholds is None:
        step = 0.05
        thresholds = torch.arange(0.5, 0.95 + 1e-5, step, dtype=torch.float32)
    recalls = torch.zeros_like(thresholds)
    precisions = torch.zeros_like(thresholds)
    # compute recall for each iou threshold
    for i, t in enumerate(thresholds):
        recalls[i] = (gt_overlaps >= t).float().sum() / float(
            num_pos
        )  # TP/(Num GT Labels) with GT being Num Images as 1 GT per image
    for i, t in enumerate(thresholds):
        precisions[i] = (gt_overlaps >= t).float().sum() / float(
            num_boxes
        )  # TP/(Num of boxes kept)
    # ar = 2 * np.trapz(recalls, thresholds)
    ar = recalls.mean()
    ap = precisions.mean()
    return {
        "ar": ar,
        "ap": ap,
        "per_source": source_outcomes,
        "recalls": recalls,
        "precisions": precisions,
        "thresholds": thresholds,
        "gt_overlaps": gt_overlaps,
        "num_pos": num_pos,
    }


def _evaluate_predictions_on_coco(coco_gt, coco_results, iou_type, kpt_oks_sigmas=None):
    """
    Evaluate the coco results using COCOEval API.
    """
    assert len(coco_results) > 0

    if iou_type == "segm":
        coco_results = copy.deepcopy(coco_results)
        # When evaluating mask AP, if the results contain bbox, cocoapi will
        # use the box area as the area of the instance, instead of the mask area.
        # This leads to a different definition of small/medium/large.
        # We remove the bbox field to let mask AP use mask area.
        for c in coco_results:
            c.pop("bbox", None)

    coco_dt = coco_gt.loadRes(coco_results)
    coco_eval = COCOeval(coco_gt, coco_dt, iou_type)
    # Use the COCO default keypoint OKS sigmas unless overrides are specified
    if kpt_oks_sigmas:
        coco_eval.params.kpt_oks_sigmas = np.array(kpt_oks_sigmas)

    if iou_type == "keypoints":
        num_keypoints = len(coco_results[0]["keypoints"]) // 3
        assert len(coco_eval.params.kpt_oks_sigmas) == num_keypoints, (
            "[COCOEvaluator] The length of cfg.TEST.KEYPOINT_OKS_SIGMAS (default: 17) "
            "must be equal to the number of keypoints. However the prediction has {} "
            "keypoints! For more information please refer to "
            "http://cocodataset.org/#keypoints-eval.".format(num_keypoints)
        )

    coco_eval.evaluate()
    coco_eval.accumulate()
    coco_eval.summarize()

    return coco_eval
